import os
from requests import HTTPError
import unittest
import datapoint


class RegionsIntegrationTestCase(unittest.TestCase):
    def setUp(self):
        self.manager = datapoint.Manager(api_key=os.environ['API_KEY'])
        self.regions = self.manager.regions

    def test_key(self):
        self.assertEqual(self.regions.api_key, os.environ['API_KEY'])

    def test_call_api(self):
        self.assertIn (
            u'RegionalFcst', self.regions.call_api('/500'))
        self.assertRaises(
            HTTPError, self.regions.call_api, '/fake_path')
        self.assertRaises(
            HTTPError, self.regions.call_api, '/500', key='fake_key')

    def test_get_all_regions(self):
        all_regions = self.regions.get_all_regions()
        sample_region = next(
            region for region in all_regions
            if region.location_id == '515')
        self.assertEqual(sample_region.name, 'UK')
        self.assertEqual(sample_region.region, 'uk')

    def test_get_raw_forecast(self):
        sample_region = self.regions.get_all_regions()[0]
        response = self.regions.get_raw_forecast(
            sample_region.location_id)['RegionalFcst']
        self.assertEqual(response['regionId'], sample_region.region)

        # Based on what Datapoint serves at time of writing...
        forecast_periods = response['FcstPeriods']['Period']
        forecast_ids = [period['id'] for period in forecast_periods]
        expected_ids = ['day1to2', 'day3to5', 'day6to15', 'day16to30']
        self.assertEqual(forecast_ids, expected_ids)



import datetime
from datapoint.exceptions import APIException


class Forecast():
    def __init__(self, frequency=""):
        self.frequency = frequency
        self.data_date = None
        self.continent = None
        self.country = None
        self.name = None
        self.longitude = None
        self.latitude = None
        self.location_id = None
        self.elevation = None
        self.days = []

    def timedelta_total_seconds(self, timedelta):
        return (
            timedelta.microseconds + 0.0 +
            (timedelta.seconds + timedelta.days * 24 * 3600) * 10 ** 6) / 10 ** 6

    def at_datetime(self, target):
        """ Return the timestep closest to the target datetime"""

        # Convert target to offset aware datetime
        if target.tzinfo is None:
            target = datetime.datetime.combine(target.date(), target.time(), self.days[0].date.tzinfo)

        num_timesteps = len(self.days[1].timesteps)
        # First check that the target is at most 1.5 hours before the first timestep
        if target < self.days[0].timesteps[0].date - datetime.timedelta(hours=1, minutes=30) and num_timesteps == 8:
            err_str = 'There is no forecast available for the requested time. ' + \
                'The requested time is more than 1.5 hours before the first available forecast'
            raise APIException(err_str)

        elif target < self.days[0].timesteps[0].date - datetime.timedelta(hours=6) and num_timesteps == 2:

            err_str = 'There is no forecast available for the requested time. ' + \
                'The requested time is more than 6 hours before the first available forecast'

            raise APIException(err_str)

        # Ensure that the target is less than 1 hour 30 minutes after the final
        # timestep.
        # Logic is correct
        # If there are 8 timesteps per day, then the target must be within 1.5
        # hours of the last timestep
        if target > (self.days[-1].timesteps[-1].date + datetime.timedelta(hours=1, minutes=30)) and num_timesteps == 8:

            err_str = 'There is no forecast available for the requested time. The requested time is more than 1.5 hours after the first available forecast'

            raise APIException(err_str)

        # If there are 2 timesteps per day, then the target must be within 6
        # hours of the last timestep
        if target > (self.days[-1].timesteps[-1].date + datetime.timedelta(hours=6)) and num_timesteps == 2:

            err_str = 'There is no forecast available for the requested time. The requested time is more than 6 hours after the first available forecast'

            raise APIException(err_str)

        # Loop over all timesteps
        # Calculate the first time difference
        prev_td = target - self.days[0].timesteps[0].date
        prev_ts = self.days[0].timesteps[0]

        for day in self.days:
            for timestep in day.timesteps:
                # Calculate the difference between the target time and the
                # timestep.
                td = target - timestep.date

                # Find the timestep which is further from the target than the
                # previous one. Return the previous timestep
                if abs(td.total_seconds()) > abs(prev_td.total_seconds()):
                    # We are further from the target
                    return prev_ts
                if abs(td.total_seconds()) < 5400 and num_timesteps == 8:
                    # if we are past the final timestep, and it is a 3 hourly
                    # forecast, check that we are within 90 minutes of it
                    return timestep
                if abs(td.total_seconds()) < 21600 and num_timesteps == 2:
                    # if we are past the final timestep, and it is a daily
                    # forecast, check that we are within 6 hours of it
                    return timestep

                prev_ts = timestep
                prev_td = td

    def now(self):
        """Function to return the closest timestep to the current time
        """

        d = datetime.datetime.now(tz=self.days[0].date.tzinfo)
        return self.at_datetime(d)

    def future(self, in_days=0, in_hours=0, in_minutes=0, in_seconds=0):
        """Return the closest timestep to a date in a given amount of time"""

        d = datetime.datetime.now(tz=self.days[0].date.tzinfo)
        target = d + datetime.timedelta(days=in_days, hours=in_hours,
                                        minutes=in_minutes, seconds=in_seconds)

        return self.at_datetime(target)



class Observation():
    def __init__(self):
        self.data_date = None
        self.continent = None
        self.country = None
        self.name = None
        self.longitude = None
        self.latitude = None
        self.location_id = None
        self.elevation = None
        # Stores a list of observations in days
        self.days = []

    def now(self):
        """
        Return the final timestep available. This is the most recent
        observation.
        """

        return self.days[-1].timesteps[-1]



from .Element import Element

class Timestep():
    def __init__(self):
        self.name = None
        self.date = None
        self.weather = None
        self.temperature = None
        self.feels_like_temperature = None
        self.wind_speed = None
        self.wind_direction = None
        self.wind_gust = None
        self.visibility = None
        self.uv = None
        self.precipitation = None
        self.humidity = None
        self.pressure = None
        self.pressure_tendency = None
        self.dew_point = None

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def elements(self):
        """Return a list of the Elements which are not None"""
        elements = [el[1] for el in self.__dict__.items() if isinstance(el[1], Element)]

        return elements

    def __str__(self):
        timestep_string = ''
        for attr, value in self.__dict__.items():
            to_append = attr + ': ' + str(value) + '\n'
            timestep_string += to_append
        return timestep_string



class Site():
    def __init__(self):
        self.name = None
        self.location_id = None
        self.elevation = None
        self.latitude = None
        self.longitude = None
        self.nationalPark = None
        self.region = None
        self.unitaryAuthArea = None

    def __str__(self):
        site_string = ''
        for attr, value in self.__dict__.items():
            to_append = attr + ': ' + str(value) + '\n'
            site_string += to_append

        return site_string



"""
Datapoint python module
"""

from datetime import datetime
from datetime import timedelta
from time import time
from math import radians, cos, sin, asin, sqrt
import pytz
from warnings import warn

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from datapoint.exceptions import APIException
from datapoint.Site import Site
from datapoint.Forecast import Forecast
from datapoint.Observation import Observation
from datapoint.Day import Day
from datapoint.Timestep import Timestep
from datapoint.Element import Element
from datapoint.regions.RegionManager import RegionManager


FORECAST_URL = "http://datapoint.metoffice.gov.uk/public/data/val/wxfcs/all/json"
OBSERVATION_URL = "http://datapoint.metoffice.gov.uk/public/data/val/wxobs/all/json"
DATE_FORMAT = "%Y-%m-%dZ"
DATA_DATE_FORMAT = "%Y-%m-%dT%XZ"
FORECAST_DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

# See:
# https://www.metoffice.gov.uk/binaries/content/assets/mohippo/pdf/3/0/datapoint_api_reference.pdf
# pages 8 onwards for a description of the response anatomy, and the elements

ELEMENTS = {
    "Day":
        {"U":"U", "V":"V", "W":"W", "T":"Dm", "S":"S", "Pp":"PPd",
        "H":"Hn", "G":"Gn", "F":"FDm", "D":"D"},
    "Night":
        {"V":"V", "W":"W", "T":"Nm", "S":"S", "Pp":"PPn",
        "H":"Hm", "G":"Gm", "F":"FNm", "D":"D"},
    "Default":
        {"V":"V", "W":"W", "T":"T", "S":"S", "Pp":"Pp",
        "H":"H", "G":"G", "F":"F", "D":"D", "U":"U"},
    "Observation":
        {"T":"T", "V":"V", "D":"D", "S":"S",
        "W":"W", "P":"P", "Pt":"Pt", "Dp":"Dp", "H":"H"}
}

WEATHER_CODES = {
    "0":"Clear night",
    "1":"Sunny day",
    "2":"Partly cloudy",
    "3":"Partly cloudy",
    "4":"Not used",
    "5":"Mist",
    "6":"Fog",
    "7":"Cloudy",
    "8":"Overcast",
    "9":"Light rain shower",
    "10":"Light rain shower",
    "11":"Drizzle",
    "12":"Light rain",
    "13":"Heavy rain shower",
    "14":"Heavy rain shower",
    "15":"Heavy rain",
    "16":"Sleet shower",
    "17":"Sleet shower",
    "18":"Sleet",
    "19":"Hail shower",
    "20":"Hail shower",
    "21":"Hail",
    "22":"Light snow shower",
    "23":"Light snow shower",
    "24":"Light snow",
    "25":"Heavy snow shower",
    "26":"Heavy snow shower",
    "27":"Heavy snow",
    "28":"Thunder shower",
    "29":"Thunder shower",
    "30":"Thunder"
}


class Manager():
    """
    Datapoint Manager object
    """

    def __init__(self, api_key=""):
        self.api_key = api_key
        self.call_response = None

        # The list of sites changes infrequently so limit to requesting it
        # every hour.
        self.forecast_sites_last_update = 0
        self.forecast_sites_last_request = None
        self.forecast_sites_update_time = 3600

        self.observation_sites_last_update = 0
        self.observation_sites_last_request = None
        self.observation_sites_update_time = 3600

        self.regions = RegionManager(self.api_key)

    def __retry_session(self, retries=10, backoff_factor=0.3,
                        status_forcelist=(500, 502, 504),
                        session=None):
        """
        Retry the connection using requests if it fails. Use this as a wrapper
        to request from datapoint
        """

        # requests.Session allows finer control, which is needed to use the
        # retrying code
        the_session = session or requests.Session()

        # The Retry object manages the actual retrying
        retry = Retry(total=retries, read=retries, connect=retries,
                      backoff_factor=backoff_factor,
                      status_forcelist=status_forcelist)

        adapter = HTTPAdapter(max_retries=retry)

        the_session.mount('http://', adapter)
        the_session.mount('https://', adapter)

        return the_session

    def __call_api(self, path, params=None, api_url=FORECAST_URL):
        """
        Call the datapoint api using the requests module

        """
        if not params:
            params = dict()
        payload = {'key': self.api_key}
        payload.update(params)
        url = "%s/%s" % (api_url, path)

        # Add a timeout to the request.
        # The value of 1 second is based on attempting 100 connections to
        # datapoint and taking ten times the mean connection time (rounded up).
        # Could expose to users in the functions which need to call the api.
        #req = requests.get(url, params=payload, timeout=1)
        # The wrapper function __retry_session returns a requests.Session
        # object. This has a .get() function like requests.get(), so the use
        # doesn't change here.

        sess = self.__retry_session()
        req = sess.get(url, params=payload, timeout=1)

        try:
            data = req.json()
        except ValueError:
            raise APIException("DataPoint has not returned any data, this could be due to an incorrect API key")
        self.call_response = data
        if req.status_code != 200:
            msg = [data[m] for m in ("message", "error_message", "status") \
                      if m in data][0]
            raise Exception(msg)
        return data

    def _distance_between_coords(self, lon1, lat1, lon2, lat2):
        """
        Calculate the great circle distance between two points
        on the earth (specified in decimal degrees).
        Haversine formula states that:

        d = 2 * r * arcsin(sqrt(sin^2((lat1 - lat2) / 2 +
        cos(lat1)cos(lat2)sin^2((lon1 - lon2) / 2))))

        where r is the radius of the sphere. This assumes the earth is spherical.
        """

        # Convert the coordinates of the points to radians.
        lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
        r = 6371

        d_hav = 2 * r * asin(sqrt((sin((lat1 - lat2) / 2))**2 + \
                                  cos(lat1) * cos(lat2) * (sin((lon1 - lon2) / 2)**2 )))

        return d_hav

    def _get_wx_units(self, params, name):
        """
        Give the Wx array returned from datapoint and an element
        name and return the units for that element.
        """
        units = ""
        for param in params:
            if str(name) == str(param['name']):
                units = param['units']
        return units

    def _weather_to_text(self, code):
        if not isinstance(code, int):
            raise ValueError("Weather code must be an integer not", type(code))
        if code < 0 or code > 30:
            raise ValueError("Weather code outof bounds, should be 0-30")
        text = WEATHER_CODES[str(code)]
        return text

    def _visibility_to_text(self, distance):
        """
        Convert observed visibility in metres to text used in forecast
        """

        if not isinstance(distance, int):
            raise ValueError("Distance must be an integer not", type(distance))
        if distance < 0:
            raise ValueError("Distance out of bounds, should be 0 or greater")

        if 0 <= distance < 1000:
            return 'VP'
        elif 1000 <= distance < 4000:
            return 'PO'
        elif 4000 <= distance < 10000:
            return 'MO'
        elif 10000 <= distance < 20000:
            return 'GO'
        elif 20000 <= distance < 40000:
            return 'VG'
        else:
            return 'EX'

    def get_all_sites(self):
        """
        Deprecated. This function returns a list of Site object.
        """
        warning_message = 'This function is deprecated. Use get_forecast_sites() instead'
        warn(warning_message, DeprecationWarning, stacklevel=2)

        return self.get_forecast_sites()

    def get_forecast_sites(self):
        """
        This function returns a list of Site object.
        """

        time_now = time()
        if (time_now - self.forecast_sites_last_update) > self.forecast_sites_update_time or self.forecast_sites_last_request is None:

            data = self.__call_api("sitelist/")
            sites = list()
            for jsoned in data['Locations']['Location']:
                site = Site()
                site.name = jsoned['name']
                site.location_id = jsoned['id']
                site.latitude = jsoned['latitude']
                site.longitude = jsoned['longitude']

                if 'region' in jsoned:
                    site.region = jsoned['region']

                if 'elevation' in jsoned:
                    site.elevation = jsoned['elevation']

                if 'unitaryAuthArea' in jsoned:
                    site.unitaryAuthArea = jsoned['unitaryAuthArea']

                if 'nationalPark' in jsoned:
                    site.nationalPark = jsoned['nationalPark']

                site.api_key = self.api_key

                sites.append(site)
            self.forecast_sites_last_request = sites
            # Only set self.sites_last_update once self.sites_last_request has
            # been set
            self.forecast_sites_last_update = time_now
        else:
            sites = self.forecast_sites_last_request

        return sites

    def get_nearest_site(self, latitude=None,  longitude=None):
        """
        Deprecated. This function returns nearest Site object to the specified
        coordinates.
        """
        warning_message = 'This function is deprecated. Use get_nearest_forecast_site() instead'
        warn(warning_message, DeprecationWarning, stacklevel=2)

        return self.get_nearest_forecast_site(latitude, longitude)

    def get_nearest_forecast_site(self, latitude,  longitude):
        """
        This function returns the nearest Site object to the specified
        coordinates.
        """

        nearest = False
        distance = None
        sites = self.get_forecast_sites()
        # Sometimes there is a TypeError exception here: sites is None
        # So, sometimes self.get_all_sites() has returned None.
        for site in sites:
            new_distance = \
                self._distance_between_coords(
                    float(site.longitude),
                    float(site.latitude),
                    float(longitude),
                    float(latitude))

            if ((distance is None) or (new_distance < distance)):
                distance = new_distance
                nearest = site

        # If the nearest site is more than 30km away, raise an error

        if distance > 30:
            raise APIException("There is no site within 30km.")

        return nearest

    def get_forecast_for_site(self, site_id, frequency="daily"):
        """
        Get a forecast for the provided site

        A frequency of "daily" will return two timesteps:
        "Day" and "Night".

        A frequency of "3hourly" will return 8 timesteps:
        0, 180, 360 ... 1260 (minutes since midnight UTC)
        """
        data = self.__call_api(site_id, {"res": frequency})
        params = data['SiteRep']['Wx']['Param']
        forecast = Forecast(frequency=frequency)

        # If the 'Location' key is missing, there is no data for the site,
        # raise an error.
        if 'Location' not in data['SiteRep']['DV']:
            err_string = ('DataPoint has not returned any data for the'
                          'requested site.')
            raise APIException(err_string)

        # Check if the other keys we need are in the data returned from the
        # datapoint API. If they are not, the data elements in the python class
        # are left as None. It is currently the responsibility of the program
        # using them to cope with this.
        if 'dataDate' in data['SiteRep']['DV']:
            forecast.data_date = datetime.strptime(data['SiteRep']['DV']['dataDate'], DATA_DATE_FORMAT).replace(tzinfo=pytz.UTC)

        if 'continent' in data['SiteRep']['DV']['Location']:
            forecast.continent = data['SiteRep']['DV']['Location']['continent']

        if 'country' in data['SiteRep']['DV']['Location']:
            forecast.country = data['SiteRep']['DV']['Location']['country']

        if 'name' in data['SiteRep']['DV']['Location']:
            forecast.name = data['SiteRep']['DV']['Location']['name']

        if 'lon' in data['SiteRep']['DV']['Location']:
            forecast.longitude = data['SiteRep']['DV']['Location']['lon']

        if 'lat' in data['SiteRep']['DV']['Location']:
            forecast.latitude = data['SiteRep']['DV']['Location']['lat']

        if 'i' in data['SiteRep']['DV']['Location']:
            forecast.location_id = data['SiteRep']['DV']['Location']['i']

        if 'elevation' in data['SiteRep']['DV']['Location']:
            forecast.elevation = data['SiteRep']['DV']['Location']['elevation']

        for day in data['SiteRep']['DV']['Location']['Period']:
            new_day = Day()
            new_day.date = datetime.strptime(day['value'], DATE_FORMAT).replace(tzinfo=pytz.UTC)

            for timestep in day['Rep']:
                new_timestep = Timestep()

                # According to the datapoint documentation
                # (https://www.metoffice.gov.uk/datapoint/product/uk-daily-site-specific-forecast),
                # the data provided are for noon (local time) and midnight
                # (local time). This implies that midnight is 00:00 and noon is
                # 12:00.

                if timestep['$'] == "Day":
                    cur_elements = ELEMENTS['Day']
                    new_timestep.date = datetime.strptime(day['value'], DATE_FORMAT).replace(tzinfo=pytz.UTC) \
                                        + timedelta(hours=12)
                elif timestep['$'] == "Night":
                    cur_elements = ELEMENTS['Night']
                    new_timestep.date = datetime.strptime(day['value'], DATE_FORMAT).replace(tzinfo=pytz.UTC)
                else:
                    cur_elements = ELEMENTS['Default']
                    new_timestep.date = datetime.strptime(day['value'], DATE_FORMAT).replace(tzinfo=pytz.UTC) \
                                        + timedelta(minutes=int(timestep['$']))

                if frequency == 'daily':
                    new_timestep.name = timestep['$']
                elif frequency == '3hourly':
                    new_timestep.name = int(timestep['$'])

                new_timestep.weather = \
                    Element(cur_elements['W'],
                            timestep[cur_elements['W']],
                            self._get_wx_units(params, cur_elements['W']))
                new_timestep.weather.text = self._weather_to_text(int(timestep[cur_elements['W']]))

                new_timestep.temperature = \
                    Element(cur_elements['T'],
                            int(timestep[cur_elements['T']]),
                            self._get_wx_units(params, cur_elements['T']))

                new_timestep.feels_like_temperature = \
                    Element(cur_elements['F'],
                            int(timestep[cur_elements['F']]),
                            self._get_wx_units(params, cur_elements['F']))

                new_timestep.wind_speed = \
                    Element(cur_elements['S'],
                            int(timestep[cur_elements['S']]),
                            self._get_wx_units(params, cur_elements['S']))

                new_timestep.wind_direction = \
                    Element(cur_elements['D'],
                            timestep[cur_elements['D']],
                            self._get_wx_units(params, cur_elements['D']))


                new_timestep.wind_gust = \
                    Element(cur_elements['G'],
                            int(timestep[cur_elements['G']]),
                            self._get_wx_units(params, cur_elements['G']))

                new_timestep.visibility = \
                    Element(cur_elements['V'],
                            timestep[cur_elements['V']],
                            self._get_wx_units(params, cur_elements['V']))

                new_timestep.precipitation = \
                    Element(cur_elements['Pp'],
                            int(timestep[cur_elements['Pp']]),
                            self._get_wx_units(params, cur_elements['Pp']))

                new_timestep.humidity = \
                    Element(cur_elements['H'],
                            int(timestep[cur_elements['H']]),
                            self._get_wx_units(params, cur_elements['H']))

                if 'U' in cur_elements and cur_elements['U'] in timestep:
                    new_timestep.uv = \
                        Element(cur_elements['U'],
                                timestep[cur_elements['U']],
                                self._get_wx_units(params, cur_elements['U']))

                new_day.timesteps.append(new_timestep)

                # The daily timesteps are not sorted by time. Sort them
                if frequency == 'daily':
                    new_day.timesteps.sort(key=lambda r: r.date)
            forecast.days.append(new_day)

        return forecast

    def get_observation_sites(self):
        """
        This function returns a list of Site objects for which observations are available.
        """
        if (time() - self.observation_sites_last_update) > self.observation_sites_update_time:
            self.observation_sites_last_update = time()
            data = self.__call_api("sitelist/", None, OBSERVATION_URL)
            sites = list()
            for jsoned in data['Locations']['Location']:
                site = Site()
                site.name = jsoned['name']
                site.location_id = jsoned['id']
                site.latitude = jsoned['latitude']
                site.longitude = jsoned['longitude']

                if 'region' in jsoned:
                    site.region = jsoned['region']

                if 'elevation' in jsoned:
                    site.elevation = jsoned['elevation']

                if 'unitaryAuthArea' in jsoned:
                    site.unitaryAuthArea = jsoned['unitaryAuthArea']

                if 'nationalPark' in jsoned:
                    site.nationalPark = jsoned['nationalPark']

                site.api_key = self.api_key

                sites.append(site)
            self.observation_sites_last_request = sites
        else:
            sites = self.observation_sites_last_request

        return sites

    def get_nearest_observation_site(self, latitude, longitude):
        """
        This function returns the nearest Site to the specified
        coordinates that supports observations
        """

        nearest = False
        distance = None
        sites = self.get_observation_sites()
        for site in sites:
            new_distance = \
                self._distance_between_coords(
                    float(site.longitude),
                    float(site.latitude),
                    float(longitude),
                    float(latitude))

            if ((distance == None) or (new_distance < distance)):
                distance = new_distance
                nearest = site

        # If the nearest site is more than 20km away, raise an error
        if distance > 20:
            raise APIException("There is no site within 30km.")

        return nearest


    def get_observations_for_site(self, site_id, frequency='hourly'):
            """
            Get observations for the provided site

            Returns hourly observations for the previous 24 hours
            """

            data = self.__call_api(site_id,{"res": frequency}, OBSERVATION_URL)

            params = data['SiteRep']['Wx']['Param']
            observation = Observation()

            # Check if keys are in data returned before using them.
            if 'dataDate' in data['SiteRep']['DV']:
                observation.data_date = datetime.strptime(data['SiteRep']['DV']['dataDate'], DATA_DATE_FORMAT).replace(tzinfo=pytz.UTC)

            if 'Location' in data['SiteRep']['DV']:
                if 'continent' in data['SiteRep']['DV']['Location']:
                    observation.continent = data['SiteRep']['DV']['Location']['continent']

                if 'country' in data['SiteRep']['DV']['Location']:
                    observation.country = data['SiteRep']['DV']['Location']['country']

                if 'name' in data['SiteRep']['DV']['Location']:
                    observation.name = data['SiteRep']['DV']['Location']['name']

                if 'lon' in data['SiteRep']['DV']['Location']:
                    observation.longitude = data['SiteRep']['DV']['Location']['lon']

                if 'lat' in data['SiteRep']['DV']['Location']:
                    observation.latitude = data['SiteRep']['DV']['Location']['lat']

                if 'i' in data['SiteRep']['DV']['Location']:
                    observation.location_id = data['SiteRep']['DV']['Location']['i']

                if 'elevation' in data['SiteRep']['DV']['Location']:
                    observation.elevation = data['SiteRep']['DV']['Location']['elevation']

            for day in data['SiteRep']['DV']['Location']['Period']:
                new_day = Day()
                new_day.date = datetime.strptime(day['value'], DATE_FORMAT).replace(tzinfo=pytz.UTC)

                # If the day only has 1 timestep, put it into a list by itself
                # so it can be treated the same as a day with multiple timesteps
                if type(day['Rep']) is not list:
                        day['Rep'] = [day['Rep']]

                for timestep in day['Rep']:
                    # As stated in
                    # https://www.metoffice.gov.uk/datapoint/product/uk-hourly-site-specific-observations,
                    # some sites do not have all parameters available for
                    # observations. The documentation does not state which
                    # fields may be absent. If the parameter is not available,
                    # nothing is returned from the API. If this happens the
                    # value of the element is set to 'Not reported'. This may
                    # change to the element not being assigned to the timestep.

                    new_timestep = Timestep()
                    # Assume the '$' field is always present.
                    new_timestep.name = int(timestep['$'])

                    cur_elements = ELEMENTS['Observation']

                    new_timestep.date = datetime.strptime(day['value'], DATE_FORMAT).replace(tzinfo=pytz.UTC) + timedelta(minutes=int(timestep['$']))

                    if cur_elements['W'] in timestep:
                        new_timestep.weather = \
                            Element(cur_elements['W'],
                                    timestep[cur_elements['W']],
                                    self._get_wx_units(params, cur_elements['W']))
                        new_timestep.weather.text = \
                            self._weather_to_text(int(timestep[cur_elements['W']]))
                    else:
                        new_timestep.weather = \
                            Element(cur_elements['W'],
                                    'Not reported')

                    if cur_elements['T'] in timestep:
                        new_timestep.temperature = \
                            Element(cur_elements['T'],
                                    float(timestep[cur_elements['T']]),
                                    self._get_wx_units(params, cur_elements['T']))
                    else:
                        new_timestep.temperature = \
                            Element(cur_elements['T'],
                                    'Not reported')

                    if 'S' in timestep:
                        new_timestep.wind_speed = \
                            Element(cur_elements['S'],
                                    int(timestep[cur_elements['S']]),
                                    self._get_wx_units(params, cur_elements['S']))
                    else:
                        new_timestep.wind_speed = \
                            Element(cur_elements['S'],
                                    'Not reported')

                    if 'D' in timestep:
                        new_timestep.wind_direction = \
                            Element(cur_elements['D'],
                                    timestep[cur_elements['D']],
                                    self._get_wx_units(params, cur_elements['D']))
                    else:
                        new_timestep.wind_direction = \
                            Element(cur_elements['D'],
                                    'Not reported')

                    if cur_elements['V'] in timestep:
                        new_timestep.visibility = \
                            Element(cur_elements['V'],
                                    int(timestep[cur_elements['V']]),
                                    self._get_wx_units(params, cur_elements['V']))
                        new_timestep.visibility.text = self._visibility_to_text(int(timestep[cur_elements['V']]))
                    else:
                        new_timestep.visibility = \
                            Element(cur_elements['V'],
                                    'Not reported')

                    if cur_elements['H'] in timestep:
                        new_timestep.humidity = \
                            Element(cur_elements['H'],
                                    float(timestep[cur_elements['H']]),
                                    self._get_wx_units(params, cur_elements['H']))
                    else:
                        new_timestep.humidity = \
                            Element(cur_elements['H'],
                                    'Not reported')

                    if cur_elements['Dp'] in timestep:
                        new_timestep.dew_point = \
                            Element(cur_elements['Dp'],
                                    float(timestep[cur_elements['Dp']]),
                                    self._get_wx_units(params,
                                                       cur_elements['Dp']))
                    else:
                        new_timestep.dew_point = \
                            Element(cur_elements['Dp'],
                                    'Not reported')

                    if cur_elements['P'] in timestep:
                        new_timestep.pressure = \
                            Element(cur_elements['P'],
                                    float(timestep[cur_elements['P']]),
                                    self._get_wx_units(params, cur_elements['P']))
                    else:
                        new_timestep.pressure = \
                            Element(cur_elements['P'],
                                    'Not reported')

                    if cur_elements['Pt'] in timestep:
                        new_timestep.pressure_tendency = \
                            Element(cur_elements['Pt'],
                                    timestep[cur_elements['Pt']],
                                    self._get_wx_units(params, cur_elements['Pt']))
                    else:
                        new_timestep.pressure_tendency = \
                            Element(cur_elements['Pt'],
                                    'Not reported')

                    new_day.timesteps.append(new_timestep)
                observation.days.append(new_day)

            return observation



import os

import appdirs


def API_profile_fname(profile_name='default'):
    """Get the API key profile filename."""
    return os.path.join(appdirs.user_data_dir('DataPoint'),
                        profile_name + '.key')


def install_API_key(api_key, profile_name='default'):
    """Put the given API key into the given profile name."""
    fname = API_profile_fname(profile_name)
    if not os.path.isdir(os.path.dirname(fname)):
        os.makedirs(os.path.dirname(fname))
    with open(fname, 'w') as fh:
        fh.write(api_key)



#!/usr/bin/env python
"""
This example will print out a simple forecast for the next 5 days.
It will allow us to explore the day, timestep and element objects.
"""

import datapoint

# Create datapoint connection
conn = datapoint.Manager(api_key="aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")

# Get nearest site and print out its name

site = conn.get_nearest_forecast_site(51.500728, -0.124626)
print(site.name)

# Get a forecast for the nearest site
forecast = conn.get_forecast_for_site(site.location_id, "3hourly")

# Loop through days and print date
for day in forecast.days:
    print("\n%s" % day.date)

    # Loop through time steps and print out info
    for timestep in day.timesteps:
        print(timestep.date)
        print(timestep.weather.text)
        print("%s%s%s" % (timestep.temperature.value,
                          '\xb0', #Unicode character for degree symbol
                          timestep.temperature.units))



class Element():
    def __init__(self, field_code=None, value=None, units=None):

        self.field_code = field_code
        self.value = value
        self.units = units

        # For elements which can also have a text value
        self.text = None

    def __str__(self):
        return str(self.value) + ' ' + str(self.units)



#!/usr/bin/env python
"""
This example checks whether it is due to rain at any point
today and then decides if we need to take an umbrella.
"""

import datapoint

# Create umbrella variable to use later
umbrella = False

# Create datapoint connection
conn = datapoint.Manager(api_key="aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")

# Get nearest site and print out its name
site = conn.get_nearest_forecast_site(51.500728, -0.124626)
print(site.name)

# Get a forecast for the nearest site
forecast = conn.get_forecast_for_site(site.location_id, "3hourly")

# Loop through all the timesteps in day 0 (today)
for timestep in forecast.days[0].timesteps:
    # Check to see if the chance of rain is more than 20% at any point
    if timestep.precipitation.value > 20:
        umbrella = True

# Print out the results
if umbrella == True:
    print("Looks like rain! Better take an umbrella.")
else:
    print("Don't worry you don't need an umbrella today.")



REGION_NAMES = {
    'os': 'Orkney & Shetland',
    'he': 'Highland & Eilean Siar',
    'gr': 'Grampian',
    'ta': 'Tayside',
    'st': 'Strathclyde',
    'dg': 'Dumfries, Galloway, Lothian',
    'ni': 'Northern Ireland',
    'yh': 'Yorkshire & the Humber',
    'ne': 'Northeast England',
    'em': 'East Midlands',
    'ee': 'East of England',
    'se': 'London & Southeast England',
    'nw': 'Northwest England',
    'wm': 'West Midlands',
    'sw': 'Southwest England',
    'wl': 'Wales',
    'uk': 'UK',
}


#!/usr/bin/env python
"""
A variation on current_weather.py which uses postcodes rather than lon lat.
"""

import datapoint
import postcodes_io_api

# Create datapoint connection
conn = datapoint.Manager(api_key="aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")


# Get longitude and latitude from postcode
postcodes_conn = postcodes_io_api.Api()
postcode = postcodes_conn.get_postcode('SW1A 2AA')
latitude = postcode['result']['latitude']
longitude = postcode['result']['longitude']

# Get nearest site and print out its name
site = conn.get_nearest_forecast_site(latitude, longitude)
print(site.name)

# Get a forecast for the nearest site
forecast = conn.get_forecast_for_site(site.location_id, "3hourly")

# Get the current timestep using now() and print out some info
now = forecast.now()
print(now.weather.text)
print("%s%s%s" % (now.temperature.value,
                  '\xb0', #Unicode character for degree symbol
                  now.temperature.units))



from datetime import datetime, date
import json
import pathlib
import requests
from requests_mock import Mocker
import unittest
from unittest.mock import patch

import datapoint

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S%z"


class MockDateTime(datetime):
	"""Replacement for datetime that can be mocked for testing."""
	def __new__(cls, *args, **kwargs):
		return datetime.__new__(datetime, *args, **kwargs)


class TestDataPoint(unittest.TestCase):
    @Mocker()
    def setUp(self, mock_request):
        with open("{}/datapoint.json".format(pathlib.Path(__file__).parent.absolute())) as f:
            mock_json = json.load(f)

        self.all_sites = json.dumps(mock_json['all_sites'])
        self.wavertree_hourly = json.dumps(mock_json['wavertree_hourly'])
        self.wavertree_daily = json.dumps(mock_json['wavertree_daily'])
        self.kingslynn_hourly = json.dumps(mock_json['kingslynn_hourly'])

        self.conn = datapoint.connection(api_key="abcdefgh-acbd-abcd-abcd-abcdefghijkl")

        mock_request.get('/public/data/val/wxfcs/all/json/sitelist/', text=self.all_sites)
        self.wavertree = self.conn.get_nearest_forecast_site(53.38374, -2.90929)
        self.kingslynn = self.conn.get_nearest_forecast_site(52.75556, 0.44231)


    @Mocker()
    @patch('datetime.datetime', MockDateTime)
    def test_wavertree_hourly(self, mock_request):
        from datetime import datetime, timezone
        MockDateTime.now = classmethod(lambda cls, **kwargs: datetime(2020, 4, 25, 12, tzinfo=timezone.utc))
        mock_request.get('/public/data/val/wxfcs/all/json/354107?res=3hourly', text=self.wavertree_hourly)

        forecast = self.conn.get_forecast_for_site(self.wavertree.location_id, "3hourly")
        now = forecast.now()

        self.assertEqual(self.wavertree.location_id, '354107')
        self.assertEqual(self.wavertree.name, 'Wavertree')

        self.assertEqual(now.date.strftime(DATETIME_FORMAT), '2020-04-25 12:00:00+0000')
        self.assertEqual(now.weather.value, '1')
        self.assertEqual(now.temperature.value, 17)
        self.assertEqual(now.feels_like_temperature.value, 14)
        self.assertEqual(now.wind_speed.value, 9)
        self.assertEqual(now.wind_direction.value, 'SSE')
        self.assertEqual(now.wind_gust.value, 16)
        self.assertEqual(now.visibility.value, 'GO')
        self.assertEqual(now.uv.value, '5')
        self.assertEqual(now.precipitation.value, 0)
        self.assertEqual(now.humidity.value, 50)

        self.assertEqual(len(forecast.days), 5)
        self.assertEqual(len(forecast.days[0].timesteps), 7)
        self.assertEqual(len(forecast.days[3].timesteps), 8)

        self.assertEqual(forecast.days[3].timesteps[7].date.strftime(DATETIME_FORMAT), '2020-04-28 21:00:00+0000')
        self.assertEqual(forecast.days[3].timesteps[7].weather.value, '7')
        self.assertEqual(forecast.days[3].timesteps[7].temperature.value, 10)
        self.assertEqual(forecast.days[3].timesteps[7].feels_like_temperature.value, 9)
        self.assertEqual(forecast.days[3].timesteps[7].wind_speed.value, 4)
        self.assertEqual(forecast.days[3].timesteps[7].wind_direction.value, 'NNE')
        self.assertEqual(forecast.days[3].timesteps[7].wind_gust.value, 11)
        self.assertEqual(forecast.days[3].timesteps[7].visibility.value, 'VG')
        self.assertEqual(forecast.days[3].timesteps[7].uv.value, '0')
        self.assertEqual(forecast.days[3].timesteps[7].precipitation.value, 9)
        self.assertEqual(forecast.days[3].timesteps[7].humidity.value, 72)


    @Mocker()
    @patch('datetime.datetime', MockDateTime)
    def test_wavertree_daily(self, mock_request):
        from datetime import datetime, timezone
        MockDateTime.now = classmethod(lambda cls, **kwargs: datetime(2020, 4, 25, 12, tzinfo=timezone.utc))
        mock_request.get('/public/data/val/wxfcs/all/json/354107?res=daily', text=self.wavertree_daily)

        forecast = self.conn.get_forecast_for_site(self.wavertree.location_id, "daily")
        now = forecast.now()

        self.assertEqual(self.wavertree.location_id, '354107')
        self.assertEqual(self.wavertree.name, 'Wavertree')

        self.assertEqual(now.date.strftime(DATETIME_FORMAT), '2020-04-25 12:00:00+0000')
        self.assertEqual(now.weather.value, '1')
        self.assertEqual(now.temperature.value, 19)
        self.assertEqual(now.feels_like_temperature.value, 18)
        self.assertEqual(now.wind_speed.value, 9)
        self.assertEqual(now.wind_direction.value, 'SSE')
        self.assertEqual(now.wind_gust.value, 16)
        self.assertEqual(now.visibility.value, 'GO')
        self.assertEqual(now.uv.value, '5')
        self.assertEqual(now.precipitation.value, 2)
        self.assertEqual(now.humidity.value, 50)

        self.assertEqual(len(forecast.days), 5)
        self.assertEqual(len(forecast.days[0].timesteps), 2)
        self.assertEqual(len(forecast.days[4].timesteps), 2)

        self.assertEqual(forecast.days[4].timesteps[1].date.strftime(DATETIME_FORMAT), '2020-04-29 12:00:00+0000')
        self.assertEqual(forecast.days[4].timesteps[1].weather.value, '12')
        self.assertEqual(forecast.days[4].timesteps[1].temperature.value, 13)
        self.assertEqual(forecast.days[4].timesteps[1].feels_like_temperature.value, 10)
        self.assertEqual(forecast.days[4].timesteps[1].wind_speed.value, 13)
        self.assertEqual(forecast.days[4].timesteps[1].wind_direction.value, 'SE')
        self.assertEqual(forecast.days[4].timesteps[1].wind_gust.value, 27)
        self.assertEqual(forecast.days[4].timesteps[1].visibility.value, 'GO')
        self.assertEqual(forecast.days[4].timesteps[1].uv.value, '3')
        self.assertEqual(forecast.days[4].timesteps[1].precipitation.value, 59)
        self.assertEqual(forecast.days[4].timesteps[1].humidity.value, 72)

    @Mocker()
    @patch('datetime.datetime', MockDateTime)
    def test_kingslynn_hourly(self, mock_request):
        from datetime import datetime, timezone
        MockDateTime.now = classmethod(lambda cls, **kwargs: datetime(2020, 4, 25, 12, tzinfo=timezone.utc))
        mock_request.get('/public/data/val/wxfcs/all/json/322380?res=3hourly', text=self.kingslynn_hourly)

        forecast = self.conn.get_forecast_for_site(self.kingslynn.location_id, "3hourly")
        now = forecast.now()

        self.assertEqual(self.kingslynn.location_id, '322380')
        self.assertEqual(self.kingslynn.name, "King's Lynn")

        self.assertEqual(now.date.strftime(DATETIME_FORMAT), '2020-04-25 12:00:00+0000')
        self.assertEqual(now.weather.value, '1')
        self.assertEqual(now.temperature.value, 14)
        self.assertEqual(now.feels_like_temperature.value, 13)
        self.assertEqual(now.wind_speed.value, 2)
        self.assertEqual(now.wind_direction.value, 'E')
        self.assertEqual(now.wind_gust.value, 7)
        self.assertEqual(now.visibility.value, 'VG')
        self.assertEqual(now.uv.value, '6')
        self.assertEqual(now.precipitation.value, 0)
        self.assertEqual(now.humidity.value, 60)

        self.assertEqual(len(forecast.days), 5)
        self.assertEqual(len(forecast.days[0].timesteps), 7)
        self.assertEqual(len(forecast.days[4].timesteps), 8)

        self.assertEqual(forecast.days[4].timesteps[5].date.strftime(DATETIME_FORMAT), '2020-04-29 15:00:00+0000')
        self.assertEqual(forecast.days[4].timesteps[5].weather.value, '12')
        self.assertEqual(forecast.days[4].timesteps[5].temperature.value, 14)
        self.assertEqual(forecast.days[4].timesteps[5].feels_like_temperature.value, 12)
        self.assertEqual(forecast.days[4].timesteps[5].wind_speed.value, 13)
        self.assertEqual(forecast.days[4].timesteps[5].wind_direction.value, 'S')
        self.assertEqual(forecast.days[4].timesteps[5].wind_gust.value, 27)
        self.assertEqual(forecast.days[4].timesteps[5].visibility.value, 'GO')
        self.assertEqual(forecast.days[4].timesteps[5].uv.value, '1')
        self.assertEqual(forecast.days[4].timesteps[5].precipitation.value, 55)
        self.assertEqual(forecast.days[4].timesteps[5].humidity.value, 68)

if __name__ == '__main__':
    unittest.main()



import datetime
import datapoint
import unittest


class TestForecast(unittest.TestCase):

    def setUp(self):
        self.forecast_3hrly = datapoint.Forecast.Forecast(frequency='3hourly')

        test_day_0 = datapoint.Day.Day()
        test_day_0.date = datetime.datetime(2020, 3, 3, tzinfo=datetime.timezone.utc)

        for i in range(5):
            ts = datapoint.Timestep.Timestep()
            ts.name = 9+(3*i)
            ts.date = datetime.datetime(2020, 3, 3, 9+(3*i), tzinfo=datetime.timezone.utc)
            test_day_0.timesteps.append(ts)

        test_day_1 = datapoint.Day.Day()
        test_day_1.date = datetime.datetime(2020, 3, 4, tzinfo=datetime.timezone.utc)

        for i in range(8):
            ts = datapoint.Timestep.Timestep()
            ts.name = 3*i
            ts.date = datetime.datetime(2020, 3, 4, 3*i, tzinfo=datetime.timezone.utc)
            test_day_1.timesteps.append(ts)

        self.forecast_3hrly.days.append(test_day_0)
        self.forecast_3hrly.days.append(test_day_1)

        test_day_0 = datapoint.Day.Day()
        test_day_0.date = datetime.datetime(2020, 3, 3, tzinfo=datetime.timezone.utc)

        for i in range(2):
            ts = datapoint.Timestep.Timestep()
            ts.name = 2*i
            ts.date = datetime.datetime(2020, 3, 3, 2*i, tzinfo=datetime.timezone.utc)
            test_day_0.timesteps.append(ts)

        test_day_1 = datapoint.Day.Day()
        test_day_1.date = datetime.datetime(2020, 3, 4, tzinfo=datetime.timezone.utc)

        for i in range(2):
            ts = datapoint.Timestep.Timestep()
            ts.name = 2*i
            ts.date = datetime.datetime(2020, 3, 4, 2*i, tzinfo=datetime.timezone.utc)
            test_day_1.timesteps.append(ts)

        self.forecast_daily = datapoint.Forecast.Forecast(frequency='daily')

        self.forecast_daily.days.append(test_day_0)
        self.forecast_daily.days.append(test_day_1)

    def test_at_datetime_1_5_hours_before_after(self):

        target_before = datetime.datetime(2020, 3, 3, 7, 0,
                                          tzinfo=datetime.timezone.utc)

        target_after = datetime.datetime(2020, 3, 4, 23, 0,
                                          tzinfo=datetime.timezone.utc)

        self.assertRaises(datapoint.exceptions.APIException,
                          self.forecast_3hrly.at_datetime, target_before)

        self.assertRaises(datapoint.exceptions.APIException,
                          self.forecast_3hrly.at_datetime, target_after)

    def test_at_datetime_6_hours_before_after(self):

        # Generate 2 timesteps These are set at 00:00 and 12:00

        target_before = datetime.datetime(2020, 3, 2, 15, 0,
                                          tzinfo=datetime.timezone.utc)

        target_after = datetime.datetime(2020, 3, 6, 7, 0,
                                          tzinfo=datetime.timezone.utc)

        self.assertRaises(datapoint.exceptions.APIException,
                          self.forecast_daily.at_datetime, target_before)

        self.assertRaises(datapoint.exceptions.APIException,
                          self.forecast_daily.at_datetime, target_after)

    def test_normal_time(self):
        target = datetime.datetime(2020, 3, 3, 10, 0,
                                          tzinfo=datetime.timezone.utc)

        nearest = self.forecast_3hrly.at_datetime(target)
        expected = datetime.datetime(2020, 3, 3, 9,
                                     tzinfo=datetime.timezone.utc)
        self.assertEqual(nearest.date, expected)

        target = datetime.datetime(2020, 3, 3, 11, 0,
                                          tzinfo=datetime.timezone.utc)

        nearest = self.forecast_3hrly.at_datetime(target)
        expected = datetime.datetime(2020, 3, 3, 12,
                                     tzinfo=datetime.timezone.utc)
        self.assertEqual(nearest.date, expected)


    def test_forecase_midnight(self):
        target = datetime.datetime(2020, 3, 4, 0, 15,
                                          tzinfo=datetime.timezone.utc)

        nearest = self.forecast_3hrly.at_datetime(target)
        expected = datetime.datetime(2020, 3, 4, 0,
                                     tzinfo=datetime.timezone.utc)
        self.assertEqual(nearest.date, expected)





"""Datapoint API to retrieve Met Office data"""

import os.path

from datapoint.Manager import Manager
import datapoint.profile


def connection(profile_name='default', api_key=None):
    """Connect to DataPoint with the given API key profile name."""
    if api_key is None:
        profile_fname = datapoint.profile.API_profile_fname(profile_name)
        if not os.path.exists(profile_fname):
            raise ValueError('Profile not found in {}. Please install your API \n'
                             'key with datapoint.profile.install_API_key('
                             '"<YOUR-KEY>")'.format(profile_fname))
        with open(profile_fname) as fh:
            api_key = fh.readlines()
    return Manager(api_key=api_key)

from . import _version
__version__ = _version.get_versions()['version']



class Day():
    def __init__(self):
        self.date = None
        self.timesteps = []

    def __str__(self):
        day_str = ''

        date_part = 'Date: ' + str(self.date) + '\n\n'
        day_str += date_part

        day_str += 'Timesteps: \n\n'
        try:
            for timestep in self.timesteps:
                day_str += str(timestep)
                day_str += '\n'

        except TypeError:
            day_str += 'No timesteps'

        return day_str



#!/usr/bin/env python
"""
This example will tell us which day would be best to hang out
our washing to dry.

We will loop over the next 5 days and decide whether it is
ok to hang out the washing. Then for the good days we will rank
them and print out the best.
"""

import datapoint
from datetime import datetime

# Set thresholds
MAX_WIND = 31 # in mph. We don't want the washing to blow away
MAX_PRECIPITATION = 20 # Max chance of rain we will accept

# Variables for later
best_day = None
best_day_score = 0 # For simplicity the score will be temperature + wind speed

# Create datapoint connection
conn = datapoint.Manager(api_key="aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")

# Get nearest site and print out its name
site = conn.get_nearest_forecast_site(51.500728, -0.124626)
print(site.name)


# Get a forecast for the nearest site
forecast = conn.get_forecast_for_site(site.location_id, "daily")

# Loop through days
for day in forecast.days:

    # Get the 'Day' timestep
    if day.timesteps[0].name == 'Day':
        timestep = day.timesteps[0]

        # If precipitation, wind speed and gust are less than their threshold
        if timestep.precipitation.value < MAX_PRECIPITATION and \
           timestep.wind_speed.value < MAX_WIND and \
           timestep.wind_gust.value < MAX_WIND:

            # Calculate the score for this timestep
            timestep_score = timestep.wind_speed.value + timestep.temperature.value

            # If this timestep scores better than the current best replace it
            if timestep_score > best_day_score:
                best_day_score = timestep_score
                best_day = day.date

# If best_day is still None then there are no good days
if best_day is None:
    print("Better use the tumble dryer")

# Otherwise print out the day
else:
    # Get the day of the week from the datetime object
    best_day_formatted = datetime.strftime(best_day, "%A")
    print("%s is the best day with a score of %s" % (best_day_formatted, best_day_score))



import datetime
import os
import unittest
import datapoint

class ManagerIntegrationTestCase(unittest.TestCase):

    def setUp(self):
        self.manager = datapoint.Manager(api_key=os.environ['API_KEY'])

    def test_site(self):
        site = self.manager.get_nearest_forecast_site(latitude=51.500728, longitude=-0.124626)
        self.assertEqual(site.name.upper(), 'HORSEGUARDS PARADE')

    def test_get_forecast_sites(self):
        sites = self.manager.get_forecast_sites()
        self.assertIsInstance(sites, list)
        # What is this assert testing
        assert sites

    def test_get_daily_forecast(self):
        site = self.manager.get_nearest_forecast_site(latitude=51.500728, longitude=-0.124626)
        forecast = self.manager.get_forecast_for_site(site.location_id, 'daily')
        self.assertIsInstance(forecast, datapoint.Forecast.Forecast)
        self.assertEqual(forecast.continent.upper(), 'EUROPE')
        self.assertEqual(forecast.country.upper(), 'ENGLAND')
        self.assertEqual(forecast.name.upper(),'HORSEGUARDS PARADE')
        self.assertLess(abs(float(forecast.latitude) - 51.500728), 0.1)
        self.assertLess(abs(float(forecast.longitude) - (-0.124626)), 0.1)
        # Forecast should have been made within last 3 hours
        tz = forecast.data_date.tzinfo
        self.assertLess(forecast.data_date - datetime.datetime.now(tz=tz),
                        datetime.timedelta(hours=3))
        # First forecast should be less than 12 hours away
        tz = forecast.days[0].timesteps[0].date.tzinfo
        self.assertLess(forecast.days[0].timesteps[0].date -
                        datetime.datetime.now(tz=tz),
                        datetime.timedelta(hours=12))
        for day in forecast.days:
            for timestep in day.timesteps:
                self.assertIn(timestep.name, ['Day', 'Night'])
                self.assertEqual(self.manager._weather_to_text(int(timestep.weather.value)),
                                 timestep.weather.text)

                self.assertGreater(timestep.temperature.value, -100)
                self.assertLess(timestep.temperature.value, 100)
                self.assertEqual(timestep.temperature.units, 'C')

                self.assertGreater(timestep.feels_like_temperature.value , -100)
                self.assertLess(timestep.feels_like_temperature.value , 100)
                self.assertEqual(timestep.feels_like_temperature.units, 'C')

                self.assertGreaterEqual(timestep.wind_speed.value, 0)
                self.assertLess(timestep.wind_speed.value, 300)
                self.assertEqual(timestep.wind_speed.units,'mph')

                for char in timestep.wind_direction.value:
                    self.assertIn(char, ['N', 'E', 'S', 'W'])
                self.assertEqual(timestep.wind_direction.units, 'compass')

                self.assertGreaterEqual(timestep.wind_gust.value, 0)
                self.assertLess(timestep.wind_gust.value, 300)
                self.assertEqual(timestep.wind_gust.units, 'mph')

                self.assertIn(timestep.visibility.value,
                    ['UN', 'VP', 'PO', 'MO', 'GO', 'VG', 'EX'])

                self.assertGreaterEqual(timestep.precipitation.value, 0)
                self.assertLessEqual(timestep.precipitation.value, 100)
                self.assertEqual(timestep.precipitation.units, '%')

                self.assertGreaterEqual(timestep.humidity.value, 0)
                self.assertLessEqual(timestep.humidity.value, 100)
                self.assertEqual(timestep.humidity.units, '%')

                if hasattr(timestep.uv, 'value'):
                    self.assertGreaterEqual(int(timestep.uv.value), 0)
                    self.assertLess(int(timestep.uv.value), 30)


    def test_get_3hour_forecast(self):
        site = self.manager.get_nearest_forecast_site(latitude=51.500728, longitude=-0.124626)
        forecast = self.manager.get_forecast_for_site(site.location_id, '3hourly')
        self.assertIsInstance(forecast, datapoint.Forecast.Forecast)
        self.assertEqual(forecast.continent.upper(), 'EUROPE')
        self.assertEqual(forecast.country.upper(), 'ENGLAND')
        self.assertEqual(forecast.name.upper(),'HORSEGUARDS PARADE')
        self.assertLess(abs(float(forecast.latitude) - 51.500728), 0.1)
        self.assertLess(abs(float(forecast.longitude) - (-0.124626)), 0.1)
        # Forecast should have been made within last 3 hours
        tz = forecast.data_date.tzinfo
        self.assertLess(forecast.data_date - datetime.datetime.now(tz=tz),
                        datetime.timedelta(hours=3))
        # First forecast should be less than 12 hours away
        tz = forecast.days[0].timesteps[0].date.tzinfo
        self.assertLess(forecast.days[0].timesteps[0].date -
                        datetime.datetime.now(tz=tz),
                        datetime.timedelta(hours=12))
        for day in forecast.days:
            for timestep in day.timesteps:
                self.assertIsInstance(timestep.name, int)
                self.assertEqual(self.manager._weather_to_text(
                    int(timestep.weather.value)), timestep.weather.text)

                self.assertGreater(timestep.temperature.value, -100)
                self.assertLess(timestep.temperature.value, 100)
                self.assertEqual(timestep.temperature.units, 'C')

                self.assertGreater(timestep.feels_like_temperature.value , -100)
                self.assertLess(timestep.feels_like_temperature.value , 100)
                self.assertEqual(timestep.feels_like_temperature.units, 'C')

                self.assertGreaterEqual(timestep.wind_speed.value, 0)
                self.assertLess(timestep.wind_speed.value, 300)
                self.assertEqual(timestep.wind_speed.units,'mph')

                for char in timestep.wind_direction.value:
                    self.assertIn(char, ['N', 'E', 'S', 'W'])
                self.assertEqual(timestep.wind_direction.units, 'compass')

                self.assertGreaterEqual(timestep.wind_gust.value, 0)
                self.assertLess(timestep.wind_gust.value, 300)
                self.assertEqual(timestep.wind_gust.units, 'mph')

                self.assertIn(timestep.visibility.value,
                    ['UN', 'VP', 'PO', 'MO', 'GO', 'VG', 'EX'])

                self.assertGreaterEqual(timestep.precipitation.value, 0)
                self.assertLessEqual(timestep.precipitation.value, 100)
                self.assertEqual(timestep.precipitation.units, '%')

                self.assertGreaterEqual(timestep.humidity.value, 0)
                self.assertLessEqual(timestep.humidity.value, 100)
                self.assertEqual(timestep.humidity.units, '%')

                if hasattr(timestep.uv, 'value'):
                    self.assertGreaterEqual(int(timestep.uv.value), 0)
                    self.assertLess(int(timestep.uv.value), 30)

    def test_get_nearest_observation_site(self):
        site = self.manager.get_nearest_observation_site(longitude=-0.1025, latitude=51.3263)
        self.assertEqual(site.name.upper(), 'KENLEY')

    def test_get_observation_sites(self):
        sites = self.manager.get_observation_sites()
        self.assertIsInstance(sites, list)
        # Mystery assert
        assert sites

    def test_get_observation_with_wind_data(self):
        observation = self.manager.get_observations_for_site(3840)
        self.assertIsInstance(observation, datapoint.Observation.Observation)
        self.assertEqual(observation.continent.upper(), 'EUROPE')
        self.assertEqual(observation.country.upper(), 'ENGLAND')
        self.assertEqual(observation.name.upper(), 'DUNKESWELL AERODROME')

        # Observation should be from within the last hour
        tz = observation.data_date.tzinfo
        self.assertLess(observation.data_date - datetime.datetime.now(tz=tz),
                        datetime.timedelta(hours=1))

        # First observation should be between 24 and 25 hours old
        tz = observation.days[0].timesteps[0].date.tzinfo
        self.assertGreater(datetime.datetime.now(tz=tz) -
                           observation.days[0].timesteps[0].date,
                           datetime.timedelta(hours=24))
        self.assertLess(datetime.datetime.now(tz=tz) -
                        observation.days[0].timesteps[0].date,
                        datetime.timedelta(hours=25))

        # Should have total 25 observations across all days
        number_of_timesteps = 0
        for day in observation.days:
            number_of_timesteps += len(day.timesteps)
        self.assertEqual(number_of_timesteps, 25)

        for day in observation.days:
            for timestep in day.timesteps:
                self.assertIsInstance(timestep.name, int)
                if timestep.weather.value != 'Not reported':
                    self.assertEqual(self.manager._weather_to_text(int(timestep.weather.value)),
                                     timestep.weather.text)
                self.assertGreater(timestep.temperature.value, -100)
                self.assertLess(timestep.temperature.value, 100)
                self.assertEqual(timestep.temperature.units, 'C')

                if timestep.wind_speed.value != 'Not reported':
                    self.assertGreaterEqual(timestep.wind_speed.value, 0)
                    self.assertLess(timestep.wind_speed.value, 300)
                    self.assertEqual(timestep.wind_speed.units,'mph')

                if timestep.wind_direction.value != 'Not reported':
                    for char in timestep.wind_direction.value:
                        self.assertIn(char, ['N', 'E', 'S', 'W'])
                        self.assertEqual(timestep.wind_direction.units, 'compass')

                self.assertGreaterEqual(timestep.visibility.value, 0)
                self.assertIn(timestep.visibility.text,
                    ['UN', 'VP', 'PO', 'MO', 'GO', 'VG', 'EX'])

                self.assertGreaterEqual(timestep.humidity.value, 0)
                self.assertLessEqual(timestep.humidity.value, 100)
                self.assertEqual(timestep.humidity.units, '%')

                self.assertGreaterEqual(timestep.dew_point.value, 0)
                self.assertLessEqual(timestep.dew_point.value, 100)
                self.assertEqual(timestep.dew_point.units, 'C')

                if timestep.pressure.value != 'Not reported':
                    self.assertGreaterEqual(timestep.pressure.value, 900)
                    self.assertLessEqual(timestep.pressure.value, 1100)
                    self.assertEqual(timestep.pressure.units, 'hpa')

                if timestep.pressure_tendency.value != 'Not reported':
                    self.assertIn(timestep.pressure_tendency.value, ['R','F','S'])
                    self.assertEqual(timestep.pressure_tendency.units, 'Pa/s')


    # def test_get_observation_without_wind_data(self):
    #     observation = self.manager.get_observations_for_site(3220)
    #     assert isinstance(observation, datapoint.Observation.Observation)
    #     assert observation.continent.upper() == 'EUROPE'
    #     assert observation.country.upper() == 'ENGLAND'
    #     assert observation.name.upper() == 'CARLISLE'

    #     # Observation should be from within the last hour
    #     tz = observation.data_date.tzinfo
    #     assert (observation.data_date
    #         - datetime.datetime.now(tz=tz) < datetime.timedelta(hours=1))

    #     # First observation should be between 24 and 25 hours old
    #     tz = observation.days[0].timesteps[0].date.tzinfo
    #     assert (datetime.datetime.now(tz=tz) - observation.days[0].timesteps[0].date > datetime.timedelta(hours=24))
    #     assert (datetime.datetime.now(tz=tz) - observation.days[0].timesteps[0].date < datetime.timedelta(hours=25))

    #     # Should have total 25 observations across all days
    #     number_of_timesteps = 0
    #     for day in observation.days:
    #         number_of_timesteps += len(day.timesteps)
    #     assert number_of_timesteps == 25

    #     for day in observation.days:
    #         for timestep in day.timesteps:
    #             assert isinstance(timestep.name, int)
    #             if timestep.weather.value != 'Not reported':
    #                 assert self.manager._weather_to_text(
    #                     int(timestep.weather.value)) == timestep.weather.text
    #             assert -100 < timestep.temperature.value < 100
    #             assert timestep.temperature.units == 'C'
    #             assert timestep.wind_speed is None
    #             assert timestep.wind_gust is None
    #             assert timestep.wind_direction is None
    #             assert 0 <= timestep.visibility.value
    #             assert (timestep.visibility.text in
    #                 ['UN', 'VP', 'PO', 'MO', 'GO', 'VG', 'EX'])
    #             assert 0 <= timestep.humidity.value <= 100
    #             assert timestep.humidity.units == '%'
    #             assert -100 < timestep.dew_point.value < 100
    #             assert timestep.dew_point.units == 'C'
    #             assert 900 < timestep.pressure.value < 1100
    #             assert timestep.pressure.units == 'hpa'
    #             assert timestep.pressure_tendency.value in ('R','F','S')
    #             assert timestep.pressure_tendency.units == 'Pa/s'



import json
from time import time

import requests

from datapoint.Site import Site
from datapoint.regions.region_names import REGION_NAMES
REGIONS_BASE_URL = 'http://datapoint.metoffice.gov.uk/public/data/txt/wxfcs/regionalforecast/json'


class RegionManager():
    '''
    Datapoint Manager for national and regional text forecasts
    '''
    def __init__(self, api_key, base_url=None):
        self.api_key = api_key
        self.all_regions_path = '/sitelist'
        if not base_url:
            self.base_url = REGIONS_BASE_URL

        # The list of regions changes infrequently so limit to requesting it
        # every hour.
        self.regions_last_update = 0
        self.regions_last_request = None
        self.regions_update_time = 3600

    def call_api(self, path, **kwargs):
        '''
        Call datapoint api
        '''
        if 'key' not in kwargs:
            kwargs['key'] = self.api_key
        req = requests.get('{0}{1}'.format(self.base_url, path), params=kwargs)

        if req.status_code != requests.codes.ok:
            req.raise_for_status()

        return req.json()

    def get_all_regions(self):
        '''
        Request a list of regions from Datapoint. Returns each Region
        as a Site object. Regions rarely change, so we cache the response
        for one hour to minimise requests to API.
        '''
        if (time() - self.regions_last_update) < self.regions_update_time:
            return self.regions_last_request

        response = self.call_api(self.all_regions_path)
        regions = []
        for location in response['Locations']['Location']:
            region = Site()
            region.location_id = location['@id']
            region.region = location['@name']
            region.name = REGION_NAMES[location['@name']]
            regions.append(region)

        self.regions_last_update = time()
        self.regions_last_request = regions
        return regions

    def get_raw_forecast(self, region_id):
        '''
        Request unformatted forecast for a specific region_id.
        '''
        return self.call_api('/{0}'.format(region_id))



#!/usr/bin/env python

from setuptools import setup
import versioneer

setup(name='datapoint',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      install_requires=[
          "requests >= 2.20.0,<3",
          "appdirs >=1,<2",
          "pytz",
      ],
      description='Python interface to the Met Office\'s Datapoint API',
      long_description_content_type='text/x-rst',
      long_description='''
Datapoint for Python
--------------------

*A Python module for accessing weather data via the Met
Office's open data API known as
`Datapoint`.*

**Disclaimer: This module is in no way part of the datapoint
project/service. This module is intended to simplify the use of
Datapoint for small Python projects (e.g school projects). No support
for this module is provided by the Met Office and may break as the
Datapoint service grows/evolves. The author will make reasonable efforts
to keep it up to date and fully featured.**

Changelog
---------

+ Explicitly state the use of semantic versioning in `README.md`.
+ Add `elements()` function to `Timestep`.
+ Remove night/day indication from weather codes which have them.
+ Change the logic used to calculate the closest timestep to a datetime. The closest timestep to the datetime is now used. Add a new function, `Forecast.at_datetime(target)` to do this. `Forecast.now()` has been changed to use this new function. The old behaviour is deprecated and available using `Forecast.now_old()`. `Forecast.future()` has been changed to use this new function. The old behaviour is deprecated and available using `Forecast.future_old()`.
+ Check if keys are returned from datapoint api in `Manager.py`. Do not attempt to read the values from the dict if they are not there.


Installation
------------

.. code:: bash

    $ pip install datapoint

You will also require a `Datapoint API`
key from http://www.metoffice.gov.uk/datapoint/API.

Features
--------

-  List forecast sites
-  Get nearest forecast site from latitiude and longitude
-  Get the following 5 day forecast types for any site
-  Daily (Two timesteps, midday and midnight UTC)
-  3 hourly (Eight timesteps, every 3 hours starting at midnight UTC)
-  Get observation sites
-  Get observations for any site

Contributing changes
--------------------

Please feel free to submit issues and pull requests.

License
-------

GPLv3.
''',
      author='Jacob Tomlinson, Emlyn Price',
      author_email='emlyn.je.price@gmail.com',
      maintainer='Emlyn Price',
      maintainer_email='emlyn.je.price@gmail.com',
      url='https://github.com/ejep/datapoint-python',
      license='GPLv3',
      packages=['datapoint', 'datapoint.regions'],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
      ]
     )



#!/usr/bin/env python
"""
This example will print out the 30 day text forecast for a region of the UK.
"""

import datapoint

# Create datapoint connection
conn = datapoint.Manager(api_key="aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")

# Get all regions and print out their details
regions = conn.regions.get_all_regions()
for region in regions:
    print((region.name, region.location_id, region.region))

# Get all forecasts for a specific region
my_region = regions[0]
forecast = conn.regions.get_raw_forecast(my_region.location_id)['RegionalFcst']

# Print the forecast details
print('Forecast for {} (issued at {}):'.format(my_region.name, forecast['issuedAt']))

sections = forecast['FcstPeriods']['Period']
for section in forecast['FcstPeriods']['Period']:
    paragraph = []
    content = section['Paragraph']

    # Some paragraphs have multiple sections
    if isinstance(content, dict):
        paragraph.append(content)
    else:
        paragraph = content

    for line in paragraph:
            print('{}\n{}\n'.format(line['title'], line['$']))



#!/usr/bin/env python
"""
This is a simple example which will print out the current weather and
temperature for our location.
"""

import datapoint

# Create datapoint connection
conn = datapoint.Manager(api_key="aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")

# Get nearest site and print out its name

site = conn.get_nearest_forecast_site(51.500728, -0.124626)
print(site.name)

# Get a forecast for the nearest site
forecast = conn.get_forecast_for_site(site.location_id, "3hourly")

# Get the current timestep using now() and print out some info
now = forecast.now()
print(now.weather.text)
print("%s%s%s" % (now.temperature.value,
                  '\xb0', #Unicode character for degree symbol
                  now.temperature.units))



import unittest
import datetime
import datapoint

class ObservationTestCase(unittest.TestCase):

    def setUp(self):
        self.observation = datapoint.Observation.Observation()

    def test_observation(self):
        # Just copy the forecast test
        test_day = datapoint.Day.Day()
        test_day.date = datetime.datetime.utcnow()

        test_timestep = datapoint.Timestep.Timestep()
        test_timestep.name = 1

        test_day.timesteps.append(test_timestep)

        self.observation.days.append(test_day)

        # What is this asserting?
        assert self.observation.now()



import unittest
import datapoint

class ManagerTestCase(unittest.TestCase):

    def setUp(self):
        self.manager = datapoint.Manager(api_key="")

    def test_weather_to_text_is_string(self):
        weather_text = self.manager._weather_to_text(0)
        self.assertIsInstance(weather_text, type(""))

    def test_weather_to_text_invalid_input_None(self):
        self.assertRaises(ValueError, self.manager._weather_to_text, None)

    def test_weather_to_text_invalid_input_out_of_bounds(self):
        self.assertRaises(ValueError, self.manager._weather_to_text, 31)

    def test_weather_to_text_invalid_input_String(self):
        self.assertRaises(ValueError, self.manager._weather_to_text, '1')

    def test_visbility_to_text_invalid_input_None(self):
        self.assertRaises(ValueError, self.manager._weather_to_text, None)

    def test_visibility_to_text_invalid_input_out_of_bounds(self):
        self.assertRaises(ValueError, self.manager._weather_to_text, -1)

    def test_visibility_to_text_invalid_input_String(self):
        self.assertRaises(ValueError, self.manager._weather_to_text, '1')
